/* Copyright (C) 2009 Registro.br. All rights reserved. 
* 
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions are 
* met:
* 1. Redistribution of source code must retain the above copyright 
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
* 
* THIS SOFTWARE IS PROVIDED BY REGISTRO.BR ``AS IS'' AND ANY EXPRESS OR
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIE OF FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
* EVENT SHALL REGISTRO.BR BE LIABLE FOR ANY DIRECT, INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
* OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
* USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
* DAMAGE.
 */
package br.registro.dnsshim.xfrd.dao.filesystem;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.security.InvalidAlgorithmParameterException;
import java.security.NoSuchAlgorithmException;
import java.security.PublicKey;
import java.security.interfaces.RSAPublicKey;
import java.security.spec.InvalidKeySpecException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

import br.registro.dnsshim.common.repository.FileSystemUtil;
import br.registro.dnsshim.domain.DnsClass;
import br.registro.dnsshim.domain.Dnskey;
import br.registro.dnsshim.domain.DnskeyAlgorithm;
import br.registro.dnsshim.domain.DnskeyFlags;
import br.registro.dnsshim.domain.DnskeyProtocol;
import br.registro.dnsshim.domain.ResourceRecord;
import br.registro.dnsshim.signer.domain.logic.DnskeyCalc;
import br.registro.dnsshim.xfrd.domain.DnskeyInfo;
import br.registro.dnsshim.xfrd.repository.DnskeyInfoRepository;

public class DnskeyInfoDao implements DnskeyInfoRepository {
	private static final Logger logger = Logger.getLogger(DnskeyInfoDao.class);
	private final String BASE_DIR;
	private static final String EXTENSION = ".key";
	
	public DnskeyInfoDao() {
		StringBuilder root = new StringBuilder();
		try {
			// {DNSSHIM_HOME|user.home}/dnsshim/xfrd/keys/SHA1#1/SHA1#2/SHA1#3/<id>.pub
			if (System.getenv("DNSSHIM_HOME") == null) {
				root.append(System.getProperty("user.home"));
				root.append(File.separatorChar);
				root.append("dnsshim");
			} else {
				root.append(System.getenv("DNSSHIM_HOME"));
			}
			root.append(File.separatorChar);
			root.append("xfrd");
			root.append(File.separatorChar);
			root.append("keys");
			root.append(File.separatorChar);

			File directory = new File(root.toString());
			FileSystemUtil.makePath(directory);

			if (logger.isDebugEnabled()) {
				logger.debug("XFRD PK (DNSKEYINFO) base dir:" + root);
			}
		} catch (IOException e) {
			logger.error(e.getMessage(), e);
		}
		BASE_DIR = root.toString();
	}

	@Override
	public void save(DnskeyInfo dnskeyInfo) throws IOException {
		StringBuilder path = new StringBuilder(BASE_DIR);
		path.append(FileSystemUtil.hash(dnskeyInfo.getIdentifier()));

		File dir = new File(path.toString());
		FileSystemUtil.makePath(dir);
		path.append(dnskeyInfo.getIdentifier()).append(EXTENSION);

		Dnskey dnskey = dnskeyInfo.getDnskey();
		File file = new File(path.toString());
		if (logger.isDebugEnabled()) {
			logger.debug("Saving public key:"+ file);
		}

		BufferedWriter bw = new BufferedWriter(new FileWriter(file));
		bw.write(dnskey.toString());
		bw.newLine();
		bw.close();
	}
	
	public void delete(String keyName) throws IOException {
		StringBuilder path = new StringBuilder(BASE_DIR);
		path.append(FileSystemUtil.hash(keyName));
		File dir = new File(path.toString());
		path.append(keyName + EXTENSION);
		File file = new File(path.toString());
		file.delete();
		
		while (dir.list().length == 0) {
			File parent = dir.getParentFile();
			dir.delete();
			dir = parent;
		}
	}

	@Override
	public DnskeyInfo findByIdentifier(String identifier) throws IOException {
		StringBuilder path = new StringBuilder(BASE_DIR);
		path.append(FileSystemUtil.hash(identifier));
		path.append(identifier);
		path.append(EXTENSION);
		File file = new File(path.toString());
		DnskeyInfo dnskeyInfo = new DnskeyInfo();
		if (file.getName().endsWith(EXTENSION)) {
			RSAPublicKey key = (RSAPublicKey) readPublicKey(file);
			dnskeyInfo.setIdentifier(identifier);
			byte[] keyDigest = DnskeyCalc.calcRsaBuffer(key);

			if (file.getName().contains("257")) {
				Dnskey dnskey = new Dnskey(ResourceRecord.APEX_OWNERNAME, DnsClass.IN, 0,
						DnskeyFlags.SEP_ON, DnskeyProtocol.PROTOCOL, DnskeyAlgorithm.RSASHA1, keyDigest);
				dnskeyInfo.setDnskey(dnskey);
			} else if (file.getName().contains("256")) {
				Dnskey dnskey = new Dnskey(ResourceRecord.APEX_OWNERNAME, DnsClass.IN, 0,
						DnskeyFlags.SEP_OFF, DnskeyProtocol.PROTOCOL, DnskeyAlgorithm.RSASHA1, keyDigest);
				dnskeyInfo.setDnskey(dnskey);
			}
		}
		return dnskeyInfo;
	}

	@Override
	public Map<String, DnskeyInfo> findAll()
			throws  IOException, InvalidAlgorithmParameterException {
		Map<String, DnskeyInfo> keys = new HashMap<String, DnskeyInfo>();
		File keysRootDirectory = new File(BASE_DIR);
		loadKeysRecursively(keysRootDirectory, keys);
		return keys;
	}

	private void loadKeysRecursively(File rootDirectory, Map<String, DnskeyInfo> keys)
		throws InvalidAlgorithmParameterException, IOException {
		File[] files = rootDirectory.listFiles();
		for (File file : files) {
			if (file.isDirectory()) {
				loadKeysRecursively(file, keys);
			} else {
				if (file.exists() && file.getName().endsWith(EXTENSION)) {
					RSAPublicKey key = (RSAPublicKey) readPublicKey(file);
					String identifier =
						file.getName().substring(0, file.getName().lastIndexOf('.'));

					DnskeyInfo dnskeyInfo = new DnskeyInfo();
					dnskeyInfo.setIdentifier(identifier);
					byte[] keyDigest = DnskeyCalc.calcRsaBuffer(key);

					if (file.getName().contains("257")) {
						Dnskey dnskey = new Dnskey(ResourceRecord.APEX_OWNERNAME, DnsClass.IN, 0,
								DnskeyFlags.SEP_ON, DnskeyProtocol.PROTOCOL, DnskeyAlgorithm.RSASHA1, keyDigest);
						dnskeyInfo.setDnskey(dnskey);
						keys.put(identifier, dnskeyInfo);
						if (logger.isDebugEnabled()) {
							logger.debug("Public Key " + identifier + " loaded");
						}

					} else if (file.getName().contains("256")) {
						Dnskey dnskey = new Dnskey(ResourceRecord.APEX_OWNERNAME, DnsClass.IN, 0,
								DnskeyFlags.SEP_OFF, DnskeyProtocol.PROTOCOL, DnskeyAlgorithm.RSASHA1, keyDigest);
						dnskeyInfo.setDnskey(dnskey);
						keys.put(identifier, dnskeyInfo);
						if (logger.isDebugEnabled()) {
							logger.debug("Public Key " + identifier + " loaded");
						}

					} else {
						logger.error("Cannot define DNSKEYFLAG (SEP_ON or SEP_OFF)");
						throw new IOException("Cannot define DNSKEYFLAG (SEP_ON or SEP_OFF)");

					}
				}
			}
		}
	}

	private PublicKey readPublicKey(File file) throws IOException {
		if (logger.isDebugEnabled()) {
			logger.debug("Reading public key:" + file);
		}
		try {
			FileReader fileReader = new FileReader(file);
			BufferedReader reader = new BufferedReader(fileReader);
			String line = reader.readLine();
			line = line.split(";")[0];
			
			String[] fields = line.split(" ");
			if (fields.length < 6) {
				throw new IOException("Invalid key file: " + file);
			}
			
			if (fields[1].equalsIgnoreCase("DNSKEY") == false) {
				throw new IOException("Invalid key file: " + file);
			}

			String keyDigest = fields[5];
			PublicKey publicKey = DnskeyCalc.getRsaPublicKey(keyDigest.getBytes());
			reader.close();
			fileReader.close();
			return publicKey;
		} catch (InvalidKeySpecException ike) {
			logger.error("cannot read pub key");
			throw new IOException("cannot read pub key");
		} catch (NoSuchAlgorithmException nsae) {
			logger.error("cannot read pub key");
			throw new IOException("cannot read pub key");
		}
	}

	@Override
	public List<String> findAllKeyIds() throws IOException {
		List<String> ids = new ArrayList<String>();
		File keysRootDirectory = new File(BASE_DIR);
		loadKeyIdsRecursively(keysRootDirectory, ids);
		return ids;
	}
	
	private void loadKeyIdsRecursively(File rootDirectory, List<String> ids) {
		File[] files = rootDirectory.listFiles();
		for (File file : files) {
			if (file.isDirectory()) {
				loadKeyIdsRecursively(file, ids);
			} else {
				if (file.exists() && file.getName().endsWith(EXTENSION)) {
					String identifier =
						file.getName().substring(0, file.getName().lastIndexOf('.'));
					ids.add(identifier);
				}
			}
		}
	}	
}
