/* Copyright (C) 2009 Registro.br. All rights reserved. 
* 
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions are 
* met:
* 1. Redistribution of source code must retain the above copyright 
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
* 
* THIS SOFTWARE IS PROVIDED BY REGISTRO.BR ``AS IS'' AND ANY EXPRESS OR
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIE OF FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
* EVENT SHALL REGISTRO.BR BE LIABLE FOR ANY DIRECT, INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
* OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
* USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
* DAMAGE.
 */
package br.registro.dnsshim.xfrd.dns.notifier.logic;

import java.util.Calendar;
import java.util.Date;
import java.util.concurrent.Callable;
import java.util.concurrent.CompletionService;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.PriorityBlockingQueue;

import org.apache.log4j.Logger;

import br.registro.dnsshim.xfrd.dns.notifier.client.NotifierClient;
import br.registro.dnsshim.xfrd.dns.notifier.domain.ZoneNotify;
import br.registro.dnsshim.xfrd.dns.protocol.DnsMessage;

public final class ZoneNotifier implements Runnable {
	private static final int MAX_RETRIES = 5;

	private static final Logger logger = Logger.getLogger(ZoneNotifier.class);

	private final ExecutorService executorService = Executors.newCachedThreadPool();
	private final CompletionService<ZoneNotify> completionService = new ExecutorCompletionService<ZoneNotify>(executorService);

	private final PriorityBlockingQueue<ZoneNotify> queue;
	

	public ZoneNotifier(PriorityBlockingQueue<ZoneNotify> queue) {
		this.queue = queue;
		Thread consumer = new Thread(new NotifierConsumer(completionService, queue));
		consumer.start();
	}

	@Override
	public void run() {
		if (logger.isInfoEnabled()) {
			logger.info("Sending notifies");
		}
		while (true) {
			try {
				if (queue.isEmpty()) {
					if (logger.isInfoEnabled()){
						logger.info("Notify queue empty. Waiting...");
					}	
					synchronized (queue) {
						queue.wait();
					}
				} else {
					ZoneNotify notify = null;
					if (!queue.isEmpty()) {
						 notify = queue.peek();
					} else {
						continue;
					}
					Date now = new Date();
					if (now.getTime() >= notify.getNextNotify().getTime()) {
						synchronized (queue) {
							notify = queue.take();
						}
					
						if (logger.isDebugEnabled()){
							logger.debug("Sending notify: " + notify);
						}
					
						completionService.submit(new Notifier(notify));

						if (notify.getNotifySent().incrementAndGet() <= MAX_RETRIES ) {
							// Put back into the queue for retry later
							Calendar calendar = Calendar.getInstance();
							calendar.setTime(notify.getNextNotify());
							calendar.add(Calendar.SECOND, 5);
							notify.setNextNotify(calendar.getTime());
							synchronized (queue) {
								queue.put(notify);
							}	
						}

					} else {
						if (logger.isDebugEnabled()) {
							logger.debug("Waiting until: " + notify.getNextNotify());
						}	
						synchronized (queue) {
							queue.wait(notify.getNextNotify().getTime() - now.getTime());
						}
					}
				}
				
			} catch (Exception e) {
				logger.error(e.getMessage(), e);
			}
		}

	}

}

class Notifier implements Callable<ZoneNotify> {
	private final ZoneNotify zoneNotify;


	public Notifier(ZoneNotify notify) {
		this.zoneNotify = notify;
	}

	@Override
	public ZoneNotify call() throws Exception {
		NotifierClient client = new NotifierClient();
		DnsMessage request = client.buildNotifyMessage(zoneNotify.getZonename());
		
		// Set the query id
		request.getPackets().get(0).getHeader().setId(zoneNotify.getQueryId());
		DnsMessage message = client.sendNotify(request, zoneNotify.getSlaveAddress(), zoneNotify.getSlavePort());
		ZoneNotify notify = new ZoneNotify();
		notify.setZonename(zoneNotify.getZonename());
		notify.setSlaveAddress(zoneNotify.getSlaveAddress());
		notify.setQueryId(message.getFirstPacket().getHeader().getId());
		return notify;
	}

}

class NotifierConsumer implements Runnable {
	private static final Logger logger = Logger.getLogger(NotifierConsumer.class);
	private final PriorityBlockingQueue<ZoneNotify> queue;
	private final CompletionService<ZoneNotify> completionService;
	public NotifierConsumer(CompletionService<ZoneNotify> completionService, PriorityBlockingQueue<ZoneNotify> queue) {
		this.completionService = completionService;
		this.queue = queue;
	}
	@Override
	public void run() {
		try {
			while (true) {
				Future<ZoneNotify> future = completionService.take();
				ZoneNotify zoneNotify = future.get();
				if (logger.isDebugEnabled()) {
					logger.debug("Got notify response from " +	zoneNotify.getSlaveAddress() +
							" for zone " + zoneNotify.getZonename());
				}
				synchronized (queue) {
					if (!queue.remove(zoneNotify)) {
						logger.error("Error while removing: " + zoneNotify);
					}
					queue.notify();
				}
			}
		} catch (Exception e) {
			logger.error(e.getMessage(), e);
		}
	}
	
}
